import scala.actors._
import java.util.Random

class Counter extends Actor {
  var director : Director = null
  var west     : West = null
  var visitors : Int = 0
  var open     : Boolean = false

  def act() {
    while (true) {
      receive {
        case "open" =>
          println("museum open for business")
          open = true
      }

      while (open) {
        receive {
          case "increment" =>
            visitors += 1
            if (visitors == 1) {
              west ! "resume"
            }
            println("New arrival: " + visitors + " customers")
          case "decrement" =>
            visitors -= 1
            if (visitors == 0) {
              west ! "pause"
            }
            println("Departure: " + visitors + " customers")
          case "close" =>
            println("museum is closing...")
            open = false
        }
      }

      while (visitors > 0) {
        receive {
          case "decrement" =>
            visitors -= 1
            println("Departure: " + visitors + " customers")
        }
      }

      director ! "all clear"
      west ! "all clear"
    }
  }
}

class East extends Actor {
  val generator : Random = new Random
  var counter : Counter = null
  var open     : Boolean = false

  def reminder() {
    val mainActor = this
    Actor.actor {
      Thread.sleep(1000+generator.nextInt(1000))
      mainActor ! "reminder"
    }
  }

  def act() {
    while (true) {

      receive {
        case "open" =>
          open = true
      }

      reminder()

      while (open) {
        receive {
          case "reminder" =>
            counter ! "increment"
            reminder()
          case "close" =>
            open = false
        }
      }
    }
  }
}

class West extends Actor {
  val generator : Random = new Random
  var counter : Counter = null
  var open      : Boolean = false
  var all_clear : Boolean = false
  var paused    : Boolean = true

  def reminder() {
    val mainActor = this
    Actor.actor {
      Thread.sleep(1000+generator.nextInt(2000))
      mainActor ! "reminder"
    }
  }

  def act() {
    while (true) {

      receive {
        case "open" =>
          open = true
      }

      reminder()

      while (open || !all_clear) {
        receive {
          case "reminder" =>
            if (!paused) { 
              counter ! "decrement"
            }
            reminder()
          case "pause" =>
            paused = true
          case "resume" =>
            paused = false
          case "close" =>
            open = false
          case "all clear" =>
             all_clear = true
        }
      }
    }
  }
}

class Director extends Actor {
  var counter   : Counter = null
  var east      : East = null
  var west      : West = null
  var all_clear : Boolean = false

  def act() {
    while (true) {
      east ! "open"
      west ! "open"
      counter ! "open"

      Thread.sleep(20000)

      east ! "close"
      west ! "close"
      counter ! "close"

      while (!all_clear) {
        receive {
          case "all clear" =>
            all_clear = true
        }
      }
    }
  }
}

object Museum {
  def main(args: Array[String]) {
    val c    = new Counter
    val east = new East
    val west = new West
    val director = new Director

    east.counter = c

    west.counter = c

    director.counter = c
    director.east = east
    director.west = west

    c.west = west
    c.director = director

    c.start()
    east.start()
    west.start()
    director.start()
  }
}
